<?php

namespace Strig\Category\Adapter;

use Strig\Category\CategoryUrlGeneratorInterface;
use Strig\Category\Entity\Category;
use Strig\Image\Image;
use Strig\Image\Sprite\Sprite;
use Zend\Db\Adapter\AdapterInterface;
use Zend\Db\TableGateway\TableGateway;
use Zend\Hydrator\NamingStrategy\MapNamingStrategy;
use Zend\Hydrator\Reflection;
use Zend\Hydrator\Strategy\ClosureStrategy;

class WordpressCategoryAdapter implements CategoryAdapterInterface
{
    protected $table;
    protected $tableMeta;
    protected $tablePosts;
    protected $spriteConfig;
    protected $urlGenerator;

    public function __construct(AdapterInterface $adapter, array $spriteConfig, CategoryUrlGeneratorInterface $urlGenerator)
    {
        $this->table = new TableGateway('wp_terms', $adapter);
        $this->tableMeta = new TableGateway('wp_termmeta', $adapter);
        $this->tablePosts = new TableGateway('wp_posts', $adapter);

        $this->spriteConfig = $spriteConfig;
        $this->urlGenerator = $urlGenerator;
    }

    /**
     * @return array
     */
    public function getAll()
    {
        $data = $this->get();
        $data = $this->arrayValuesRecursive($data);

        return array_values($data);
    }


    /**
     * @param $parent
     * @return array
     */
    public function getByParent($parent)
    {
        $data = $this->get($parent);
        $data = $this->arrayValuesRecursive($data);

        return array_values($data);
    }


    /**
     * @param array $categories
     * @param array $dimensions
     * @param string $path
     * @param string $baseUri
     * @return array
     * @throws \ImagickException
     */
    public function getSprites(array $categories, array $dimensions, string $path, string $baseUri)
    {
        $spriteName = md5($path);

        // Add new dimension
        // $this->spriteConfig['dimensions'][] = $dimension;

        $images = $this->getImageList($categories);
        $data = $this->makeSpriteMap($spriteName, $images, $dimensions, $baseUri);

        return $data;
    }


    /**
     * @param array $categories
     * @param string $dimension
     * @param string $path
     * @param string $baseUri
     * @return array
     */
    public function getImages(array $categories, string $dimension, string $path, string $baseUri)
    {
        $data = [];

        $images = $this->getImageList($categories);

        foreach ($images as $id => $image) {
            $image = $image
                ? $this->makeImage($image, $dimension)
                : null;

            $data[] = [
                'id' => $id,
                'image' => $image
            ];
        }

        return $data;
    }


    /**
     * @param array $elements
     * @param null $parent
     * @return array
     */
    protected function buildTree(array &$elements, $parent = null)
    {
        $parent = $parent ?? 0;
        $branch = [];

        foreach ($elements as $element) {
            if ($element['parentId'] == $parent) {
                $children = $this->buildTree($elements, $element['id']);
                if ($children) {
                    $element['children'] = $children;
                } else {
                    $element['children'] = [];
                }
                $branch[$element['id']] = $element;
                unset($elements[$element['id']]);
            }
        }
        return $branch;
    }


    /**
     * @param array $array
     * @return mixed
     */
    protected function arrayValuesRecursive($array)
    {
        foreach ($array as $key => $value) {
            if (is_array($value)) {
                $array[$key] = $this->arrayValuesRecursive($value);
            }
        }

        if (isset($array['children'])) {
            $array['children'] = array_values($array['children']);
        }

        return $array;
    }

    /**
     * @param array $array
     * @param string $key
     * @return array
     */
    protected function uniqueMultidim(array $array, string $key)
    {
        $temp_array = [];
        $i = 0;
        $key_array = [];

        foreach ($array as $val) {
            if (!in_array($val[$key], $key_array)) {
                $key_array[$i] = $val[$key];
                $temp_array[$i] = $val;
            }
            $i++;
        }
        return $temp_array;
    }


    protected function get($parent = null, $flat = true)
    {
        $parent = $parent ?: null;
        $data = [];

        $columns = [
            'term_id as id',
            'name',
            'slug'
        ];

        $joinColumns = [
            'description',
            'parent'
        ];

        $select = $this->table->getSql()->select();
        $select->columns($columns);
        $select->join('wp_term_taxonomy', 'wp_term_taxonomy.term_id = wp_terms.term_id', $joinColumns, 'left');
        $select->where('wp_term_taxonomy.taxonomy = "product_cat"');

        if (isset($parent)) {
            $select->where('wp_term_taxonomy.parent  = ' . $parent);
        }

        $resultSet = $this->table->selectWith($select);

        // Get metas
        $metaIds = [];
        foreach ($resultSet as $row) {
            $metaIds[] = $row['id'];
            $data[$row['id']] = $row;
        }

        if (!empty($metaIds)) {
            $metaSelect = $this->tableMeta->getSql()->select();
            $metaSelect->where->in('term_id', $metaIds);
            $metaResultSet = $this->tableMeta->selectWith($metaSelect);

            foreach ($metaResultSet as $metaRow) {
                $data[$metaRow['term_id']][$metaRow['meta_key']] = $metaRow['meta_value'];
            }
        }


        $namingStrategyArray = [
            'parent' => 'parentId',
            '_sale_badge_text' => 'saleBadgeText',
            '_sale_badge_size' => 'saleBadgeSize',
            '_upsale_tecdoc_artid' => 'upsaleTecdocArtId',
            '_upsale_artcodes' => 'upsaleArtCodes'
        ];

        $hydrator = new Reflection();
        $hydrator->setNamingStrategy(new MapNamingStrategy($namingStrategyArray));

        try {
            $unescapeStrategy = new ClosureStrategy(null, function ($value) {
                return html_entity_decode($value);
            });

            $customExplodeStrategy = new ClosureStrategy(null, function ($value) {
                return array_map('intval', array_filter(explode(';', $value)));
            });

            $castIntStrategy = new ClosureStrategy(null, function ($value) {
                return (int)$value;
            });

            $hydrator->addStrategy('id', $castIntStrategy);
            $hydrator->addStrategy('parentId', $castIntStrategy);
            $hydrator->addStrategy('order', $castIntStrategy);

            $hydrator->addStrategy('name', $unescapeStrategy);
            $hydrator->addStrategy('description', $unescapeStrategy);

            $hydrator->addStrategy('genericArticleIds', $customExplodeStrategy);
            $hydrator->addStrategy('genericArticleMotoIds', $customExplodeStrategy);
            $hydrator->addStrategy('upsaleTecdocArtId', $customExplodeStrategy);
            $hydrator->addStrategy('upsaleArtCodes', $customExplodeStrategy);
        } catch (\Exception $e) {
        }

        $extractor = new Reflection();

        $tree = [];
        $generator = $this->urlGenerator;
        foreach ($data as $datum) {
            /** @var Category $category */
            $category = $hydrator->hydrate((array)$datum, new Category());
            $category->setUrl($generator($category));
            $tree[] = $extractor->extract($category);
        }

        return $flat ?
            $tree
            : $this->buildTree($tree, $parent);
    }

    /**
     * @param Sprite $spriteMap
     * @param string $name
     * @return string|array
     */
    protected function makeSprite(Sprite $spriteMap, string $name)
    {
        $sprite = (string)null;

        try {
            $dimensions = $this->spriteConfig['dimensions'];
            foreach ($dimensions as $dimension) {
                list($width, $height) = explode('x', $dimension);
                $sprite = $spriteMap->render($name, $width, $height);
            }
        } catch (\ImagickException $e) {
            var_dump($e);
        }

        return $sprite;
    }

    /**
     * @param string $name
     * @param array $images
     * @param array $dimensions
     * @param string $baseUri
     * @return array
     * @throws \ImagickException
     */
    protected function makeSpriteMap(string $name, array $images, array $dimensions, string $baseUri)
    {
        $spriteMap = new Sprite($name, $this->spriteConfig);
        $data = [];
        $maps = [];

        foreach ($dimensions as $dimension) {
            list($width, $height) = explode('x', $dimension);

            $maps[] = [
                'size' => $dimension,
                'url' => rtrim($baseUri, '/') . '/' . ltrim($spriteMap->getImageUrl($width, $height), '/')
            ];

            foreach ($images as $id => $image) {
                if (!$image) {
                    $index = -1;
                } else {
                    $sprite = $spriteMap->render($image, $width, $height);
                    list($x, $y, $index) = explode(' ', $sprite);
                }


                $data[] = [
                    'id' => $id,
                    'index' => (int)$index
                ];
            }
        }

        $spriteMap->flush();

        return [
            'maps' => $maps,
            'images' => $this->uniqueMultidim($data, 'id')
        ];
    }


    /**
     * @param string $image
     * @param string $dimension
     * @return string
     */
    protected function makeImage(string $image, string $dimension)
    {
        list($width, $height) = explode('x', $dimension);

        $baseDir = $this->spriteConfig['baseDir'];
        $outputDir = $this->spriteConfig['outputDir'];

        $image = new Image($baseDir . '/' . $image);
        $image->resize($width, $height);
        return $image->save($outputDir);

//        return $image->getImagePath();
    }


    /**
     * @param array $categories
     * @return array
     */
    protected function getImageList(array $categories)
    {
        $images = [];
        foreach ($categories as $category) {
            $images[$category] = null;
        }

        $select = $this->table->getSql()->select();
        $select->columns(['term_id']);
        $select->where->in('term_id', $categories);
        $resultSet = $this->table->selectWith($select);

        if ($resultSet->count() < 1) return $images;

        $metaIds = [];
        foreach ($resultSet as $row) {
            $metaIds[] = $row['term_id'];
        }

        $metaSelect = $this->tableMeta->getSql()->select();
        $metaSelect->columns(['term_id', 'meta_value']);
        $metaSelect->where('meta_key = "thumbnail_id"');
        $metaSelect->where->in('term_id', $metaIds);
        $metaResultSet = $this->tableMeta->selectWith($metaSelect);

//        var_dump($metaResultSet->next()->current());
        if ($metaResultSet->count() < 1) return $images;

        $thumbnailIds = [];
        foreach ($metaResultSet as $metaRow) {
            $thumbnailIds[] = $metaRow['meta_value'];
            $images[$metaRow['term_id']] = $metaRow['meta_value'];
        }

        $thumbnailSelect = $this->tablePosts->getSql()->select();
        $thumbnailSelect->columns(['ID', 'post_title']);
        $thumbnailSelect->where->in('id', $thumbnailIds);
        $thumbnailResultSet = $this->tablePosts->selectWith($thumbnailSelect);

        foreach ($thumbnailResultSet as $thumbnailRow) {
            $images[array_search($thumbnailRow['ID'], $images)] = $thumbnailRow['post_title'];
        }

        return $images;
    }
}
